unit Exe32Image;

interface

uses Windows, SysUtils, Classes, Dialogs;

type
    PSection = ^TSection;
    TSection = record
        Name: array [0..7] of Char;  { $00 name of this section }
        VirtualSize: LongInt;        { $08 size of code or data before rounding }
        VirtualAddress: LongInt;     { $0C RVA to where loader should map section }
        SizeOfRawData: LongInt;      { $10 size of code or data after rounding }
        PointerToRawData: LongInt;   { $14 file-based offset to raw data }
        PointerToRelocs: LongInt;    { $18 file-based offset to relocation info }
        PointerToLineNums: LongInt;  { $1C file-based offset to line number table }
        NumRelocs: Word;             { $20 1number of relocations in table }
        NumLineNums: Word;           { $22 number of line numbers in table }
        Flags: LongInt;              { $24 flags for this section }
    end;

    PByteArray = ^ByteArray;
    PWordArray = ^WordArray;
    PLongArray = ^LongArray;
    ByteArray = array [0..1000000] of Byte;
    WordArray = array [0..1000000] of Word;
    LongArray = array [0..1000000] of LongInt;

    Directory = record
        VirtualAddress: LongInt;
        Size: LongInt;
    end;

    PEHeader = record
        Sig: LongInt;                             { $00 }
        Machine: Word;                            { $04 }
        NumSections: Word;                        { $06 }
        TimeDataStamp: LongInt;                   { $08 }
        PointerToSymbolTable: LongInt;            { $0C }
        NumberOfSymbols: LongInt;                 { $10 }
        SizeOfOptionalHeader: Word;               { $14 }
        Characteristics: Word;                    { $16 }
        Magic: Word;                              { $18 }
        MajorLinkerVersion: Byte;                 { $1A }
        MinorLinkerVersion: Byte;                 { $1B }
        SizeOfCode: LongInt;                      { $1C }
        SizeOfInitializedData: LongInt;           { $20 }
        SizeOfUninitializedData: LongInt;         { $24 }
        AddressOfEntryPoint: LongInt;             { $28 }
        BaseOfCode: LongInt;                      { $2C }
        BaseOfData: LongInt;                      { $30 }
        ImageBase: LongInt;                       { $34 }
        SectionAlignment: LongInt;                { $38 }
        FileAlignment: LongInt;                   { $3C }
        MajorOperatingSystemVersion: Word;        { $40 }
        MinorOperatingSystemVersion: Word;        { $42 }
        MajorImageversion: Word;                  { $44 }
        MinorImageversion: Word;                  { $46 }
        MajorSubsystemVersion: Word;              { $48 }
        MinorSubsystemVersion: Word;              { $4A }
        Reserved1: LongInt;                       { $4C }
        SizeOfImage: LongInt;                     { $50 }
        SizeOfHeaders: LongInt;                   { $54 }
        CheckSum: LongInt;                        { $58 }
        Subsystem: Word;                          { $5C }
        DLLCharacteristics: Word;                 { $5E }
        SizeOfStackReserve: LongInt;              { $60 }
        SizeOfStackCommit: LongInt;               { $64 }
        SizeOfHeapReserve: LongInt;               { $68 }
        SizeOfHeapCommit: LongInt;                { $6C }
        LoaderFlags: LongInt;                     { $70 }
        NumberOfRvaAndSizes: LongInt;             { $74 }
        ImageDirs: array [0..15] of Directory;    { $78 }
    end;

    EExeImage = class (Exception);

    TExeImage = class (TComponent)
    private
        FileName: String;
        Imports: TStringList;
        FileHandle: THandle;
        FileMapping: THandle;
        ImageB: PByteArray;
        ImageW: PWordArray;
        ImageL: PLongArray;
        PEImage: ^PEHeader;
        function FindSection (const SectionName: String; var Section: PSection): Boolean;
        function RVAToAddress (RVA: LongInt; Section: PSection): PChar;
        procedure ProcessImports;
    public
        constructor OpenImage (AOwner: TComponent; const FName: String);
        destructor Destroy; override;
        property ImportNames: TStringList read Imports;
    end;

implementation

procedure Panic (const ErrMsg: ShortString);
begin
    Raise EExeImage.Create (ErrMsg);
end;

destructor TExeImage.Destroy;
begin
    if FileHandle <> -1 then
    begin
        UnmapViewOfFile (ImageB);
        CloseHandle (FileMapping);
        CloseHandle (FileHandle);
    end;

    ImportNames.Free;
    Inherited Destroy;
end;

constructor TExeImage.OpenImage (AOwner: TComponent; const FName: String);
begin
    Inherited Create (AOwner);
    FileName := FName;
    FileHandle := CreateFile (PChar (FName), Generic_Read, File_Share_Read, Nil, Open_Existing, File_Attribute_Normal, 0);
    if FileHandle = -1 then Panic ('Couldn''t open: ' + FName);

    { Create the file mapping }
    FileMapping := CreateFileMapping (FileHandle, Nil, Page_ReadOnly, 0, 0, Nil);
    if FileMapping = 0 then Panic ('CreateFileMapping returned Nil');
    ImageB := MapViewOfFile (FileMapping, File_Map_Read, 0, 0, 0);
    if ImageB = Nil then Panic ('MapViewOfFile returned Nil');

    { Validate the file as a PE executable }
    ImageW := Pointer (ImageB);
    ImageL := Pointer (ImageB);
    if ImageW^[0] <> $5A4D then Panic ('Not an executable file');

    PEImage := Pointer (ImageL^[$F] + PChar (ImageW));
    if IsBadReadPtr (PEImage, sizeof (PEHeader)) or (PEImage^.Sig <> $00004550) then
        Panic ('Not a PE (WIN32 Executable) file');

    Imports := TStringList.Create;      Imports.Sorted := True;
    ProcessImports;
end;

function TExeImage.RVAToAddress (RVA: LongInt; Section: PSection): PChar;
begin
    Result := Pointer (ImageB);
    if Section = Nil then Inc (Result, RVA) else with Section^ do
        Inc (Result, RVA + PointerToRawData - VirtualAddress);
end;

function TExeImage.FindSection (const SectionName: String; var Section: PSection): Boolean;
var
    Count: Integer;
begin
    Result := True;
    Section := Pointer (PChar (PEImage) + sizeof (PEHeader));

    Count := PEImage^.NumSections;
    while Count <> 0 do begin
        if StrLIComp (Section^.Name, PChar (SectionName), 8) = 0 then Exit;
        Inc (Section);
        Dec (Count);
    end;

    Result := False;
end;

procedure TExeImage.ProcessImports;
type
    ImageImportDescriptor = record
        Hints: LongInt;              { $00 offset of hint import information }
        TimeDateStamp: LongInt;      { $04 time/date stamp of imported file  }
        ForwarderChain: LongInt;     { $08 for DLL call forwarding           }
        Name: LongInt;               { $0C RVA to name of this imported DLL  }
        FirstThunk: LongInt;         { $10 RVA to image_thunk_data info      }
    end;
var
    Section: PSection;
    pIID: ^ImageImportDescriptor;
    ModName: String;
    Thunk: PLongInt;
    Start, Stop: Integer;
    TheName: PChar;

    function FixupImportName (const RawName: PChar): String;
    var p: PChar;
    begin
        Result := RawName;
        p := StrRScan (RawName, '@');
        if p <> Nil then SetLength (Result, p - RawName);
    end;

begin
    if not FindSection ('.idata', Section) then Exit;
    pIID := Pointer (PChar (ImageW) + Section.PointerToRawData);
    while (pIID^.TimeDateStamp <> 0) or (pIID^.Name <> 0) do
    begin
        { Step through each Image Import Descriptor in turn }
        ModName := RVAToAddress (pIID^.Name, Section);
        Thunk := Pointer (RVAToAddress (pIID^.FirstThunk, Section));
        // If the pointer that thunk points to is outside of the .idata
	// section, it looks like this file is "pre-fixed up" with regards
	// to the thunk table.  In this situation, we'll need to fall back
	// to the hint-name (aka, the "Characteristics") table.
	Start := Section^.VirtualAddress;
        Stop  := Start + Section^.SizeOfRawData;
        if (Thunk^ <= Start) or (Thunk^ >= Stop) then
	begin
	    if pIID^.Hints = 0 then Exit;
            Thunk := Pointer (pIID^.Hints);
  	    if (LongInt (Thunk) <= Start) or (LongInt (Thunk) >= Stop) then Exit;
  	    Thunk := Pointer (RVAToAddress (LongInt (Thunk), Section));
        end;

        while Thunk^ <> 0 do begin
            if (Thunk^ and $80000000) = 0 then begin
                { Import by name }
                TheName := RVAToAddress (Thunk^, Section);
                ImportNames.Add (ModName + '|' + FixupImportName (TheName + 2));
            end else ImportNames.Add (ModName + '|' + IntToStr (Thunk^ and $7fffffff));
            Inc (Thunk);
        end;

        Inc (pIID);
    end;
end;

end.



